var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createWebFileSystemHost = void 0;
const path = require("typesafe-path");
const vscode_html_languageservice_1 = require("vscode-html-languageservice");
const vscode = require("vscode-languageserver");
const vscode_uri_1 = require("vscode-uri");
const protocol_1 = require("../protocol");
const utilities_1 = require("./ts/utilities");
const uriMap_1 = require("./uriMap");
const shared = require("@volar/shared");
let currentCwd = '/';
function createWebFileSystemHost() {
    const instances = (0, uriMap_1.createUriMap)();
    const onDidChangeWatchedFilesCb = new Set();
    const root = {
        dirs: new Map(),
        fileTexts: new Map(),
        fileTypes: new Map(),
        searched: false,
    };
    const pendings = [];
    const changes = [];
    const onReadys = [];
    let loading = false;
    let connection;
    return {
        ready(_connection) {
            connection = _connection;
            connection.onDidChangeWatchedFiles(params => {
                for (const change of params.changes) {
                    const fsPath = shared.getPathOfUri(change.uri);
                    const dir = getDir(path.dirname(fsPath));
                    const name = path.basename(fsPath);
                    if (change.type === vscode.FileChangeType.Created) {
                        dir.fileTypes.set(name, vscode_html_languageservice_1.FileType.File);
                    }
                    else if (change.type === vscode.FileChangeType.Changed) {
                        dir.fileTexts.delete(name);
                    }
                    else {
                        dir.fileTypes.delete(name);
                        dir.fileTexts.delete(name);
                    }
                }
                fireChanges(params);
            });
            for (const cb of onReadys) {
                cb(connection);
            }
            onReadys.length = 0;
        },
        clearCache() {
            root.dirs.clear();
            root.fileTexts.clear();
            root.fileTypes.clear();
            root.searched = false;
        },
        getWorkspaceFileSystem(rootUri) {
            let sys = instances.uriGet(rootUri.toString());
            if (!sys) {
                sys = createWorkspaceFileSystem(rootUri);
                instances.uriSet(rootUri.toString(), sys);
            }
            return sys;
        },
        onDidChangeWatchedFiles: cb => {
            onDidChangeWatchedFilesCb.add(cb);
            return () => onDidChangeWatchedFilesCb.delete(cb);
        },
    };
    function createWorkspaceFileSystem(rootUri) {
        const rootPath = shared.getPathOfUri(rootUri.toString());
        return {
            newLine: '\n',
            useCaseSensitiveFileNames: false,
            getCurrentDirectory: () => shared.getPathOfUri(rootUri.toString()),
            fileExists,
            readFile,
            readDirectory,
            getDirectories,
            resolvePath,
            realpath: path => path, // TODO: cannot implement with vscode
        };
        function resolvePath(fsPath) {
            if (currentCwd !== rootPath) {
                process.chdir(rootPath);
                currentCwd = rootPath;
            }
            return path.resolve(fsPath);
        }
        function fileExists(fsPath) {
            fsPath = resolvePath(fsPath);
            const dir = getDir(path.dirname(fsPath));
            const name = path.basename(fsPath);
            if (dir.fileTypes.has(name)) {
                return dir.fileTypes.get(name) === vscode_html_languageservice_1.FileType.File || dir.fileTypes.get(name) === vscode_html_languageservice_1.FileType.SymbolicLink;
            }
            dir.fileTypes.set(name, undefined);
            if (connection) {
                addPending('Stat', fsPath, statAsync(connection, fsPath, dir));
            }
            else {
                onReadys.push((connection) => addPending('Stat', fsPath, statAsync(connection, fsPath, dir)));
            }
            return false;
        }
        function readFile(fsPath) {
            fsPath = resolvePath(fsPath);
            const dir = getDir(path.dirname(fsPath));
            const name = path.basename(fsPath);
            if (dir.fileTexts.has(name)) {
                return dir.fileTexts.get(name);
            }
            dir.fileTexts.set(name, '');
            if (connection) {
                addPending('Read File', fsPath, readFileAsync(connection, fsPath, dir));
            }
            else {
                onReadys.push((connection) => addPending('Read File', fsPath, readFileAsync(connection, fsPath, dir)));
            }
            return '';
        }
        function readDirectory(fsPath, extensions, exclude, include, depth) {
            fsPath = resolvePath(fsPath);
            return (0, utilities_1.matchFiles)(fsPath, extensions, exclude, include, false, rootPath, depth, _dirPath => {
                let dirPath = _dirPath;
                dirPath = resolvePath(dirPath);
                const dir = getDir(dirPath);
                const files = [...dir.fileTypes];
                if (!dir.searched) {
                    dir.searched = true;
                    if (connection) {
                        addPending('Read Directory', dirPath, readDirectoryAsync(connection, dirPath, dir));
                    }
                    else {
                        onReadys.push((connection) => addPending('Read Directory', dirPath, readDirectoryAsync(connection, dirPath, dir)));
                    }
                }
                return {
                    files: files.filter(file => file[1] === vscode_html_languageservice_1.FileType.File).map(file => file[0]),
                    directories: files.filter(file => file[1] === vscode_html_languageservice_1.FileType.Directory).map(file => file[0]),
                };
            }, path => path);
        }
        // for import path completion
        function getDirectories(fsPath) {
            fsPath = resolvePath(fsPath);
            const dir = getDir(fsPath);
            const files = [...dir.fileTypes];
            if (!dir.searched) {
                dir.searched = true;
                if (connection) {
                    addPending('Read Directory', fsPath, readDirectoryAsync(connection, fsPath, dir));
                }
                else {
                    onReadys.push((connection) => addPending('Read Directory', fsPath, readDirectoryAsync(connection, fsPath, dir)));
                }
            }
            return files.filter(file => file[1] === vscode_html_languageservice_1.FileType.Directory).map(file => file[0]);
        }
        function statAsync(connection, fsPath, dir) {
            return __awaiter(this, void 0, void 0, function* () {
                const uri = shared.getUriByPath(fsPath);
                if (uri.startsWith('https://cdn.jsdelivr.net/npm/')) { // stat request always response file type for jsdelivr
                    const text = yield readJsdelivrFile(connection, uri);
                    if (text !== undefined) {
                        const name = path.basename(fsPath);
                        dir.fileTypes.set(name, vscode_html_languageservice_1.FileType.File);
                        dir.fileTexts.set(name, text);
                        changes.push({
                            uri: uri,
                            type: vscode.FileChangeType.Created,
                        });
                    }
                }
                else {
                    const result = yield connection.sendRequest(protocol_1.FsStatRequest.type, uri);
                    if ((result === null || result === void 0 ? void 0 : result.type) === vscode_html_languageservice_1.FileType.File || (result === null || result === void 0 ? void 0 : result.type) === vscode_html_languageservice_1.FileType.SymbolicLink) {
                        const name = path.basename(fsPath);
                        dir.fileTypes.set(name, result.type);
                        changes.push({
                            uri: uri,
                            type: vscode.FileChangeType.Created,
                        });
                    }
                }
            });
        }
        function readFileAsync(connection, fsPath, dir) {
            return __awaiter(this, void 0, void 0, function* () {
                const uri = shared.getUriByPath(fsPath);
                if (uri.startsWith('https://cdn.jsdelivr.net/npm/')) {
                    const text = yield readJsdelivrFile(connection, uri);
                    if (text !== undefined) {
                        const name = path.basename(fsPath);
                        dir.fileTexts.set(name, text);
                        changes.push({
                            uri: uri,
                            type: vscode.FileChangeType.Changed,
                        });
                    }
                }
                else {
                    const text = yield connection.sendRequest(protocol_1.FsReadFileRequest.type, uri);
                    if (text) {
                        const name = path.basename(fsPath);
                        dir.fileTexts.set(name, text);
                        changes.push({
                            uri: uri,
                            type: vscode.FileChangeType.Changed,
                        });
                    }
                }
            });
        }
        function readJsdelivrFile(connection, uri) {
            var _a;
            return __awaiter(this, void 0, void 0, function* () {
                // ignore .js because it's no help for intellisense
                if (uri.endsWith('.d.ts') || uri.endsWith('.json')) {
                    const text = (_a = yield connection.sendRequest(protocol_1.FsReadFileRequest.type, uri)) !== null && _a !== void 0 ? _a : undefined;
                    if (text !== undefined
                        // ignore https://cdn.jsdelivr.net/npm/@vue/runtime-dom
                        && text.indexOf('Minified by jsDelivr') === -1) {
                        return text;
                    }
                }
            });
        }
        function readDirectoryAsync(connection, fsPath, dir) {
            return __awaiter(this, void 0, void 0, function* () {
                const uri = shared.getUriByPath(fsPath);
                const result = yield connection.sendRequest(protocol_1.FsReadDirectoryRequest.type, uri);
                for (const [name, fileType] of result) {
                    if (dir.fileTypes.get(name) !== fileType && (fileType === vscode_html_languageservice_1.FileType.File || fileType === vscode_html_languageservice_1.FileType.SymbolicLink)) {
                        changes.push({
                            uri: shared.getUriByPath(path.join(fsPath, name)),
                            type: vscode.FileChangeType.Created,
                        });
                    }
                    dir.fileTypes.set(name, fileType);
                }
            });
        }
    }
    function addPending(action, fileName, p) {
        return __awaiter(this, void 0, void 0, function* () {
            pendings.push([action, fileName, p]);
            if (loading === false) {
                loading = true;
                const progress = yield (connection === null || connection === void 0 ? void 0 : connection.window.createWorkDoneProgress());
                progress === null || progress === void 0 ? void 0 : progress.begin(action);
                let i = 0;
                while (pendings.length) {
                    const current = pendings.shift();
                    progress === null || progress === void 0 ? void 0 : progress.report(i / pendings.length, vscode_uri_1.URI.parse(shared.getUriByPath(current[1])).fsPath);
                    yield current;
                    i++;
                }
                progress === null || progress === void 0 ? void 0 : progress.done();
                if (changes.length) {
                    fireChanges({ changes: [...changes] });
                    changes.length = 0;
                }
                loading = false;
            }
        });
    }
    function fireChanges(params) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const cb of [...onDidChangeWatchedFilesCb]) {
                if (onDidChangeWatchedFilesCb.has(cb)) {
                    yield cb(params);
                }
            }
        });
    }
    function getDir(dirPath) {
        const dirNames = [];
        let currentDirPath = dirPath;
        let currentDirName = path.basename(currentDirPath);
        while (currentDirName !== '') {
            dirNames.push(currentDirName);
            currentDirPath = path.dirname(currentDirPath);
            currentDirName = path.basename(currentDirPath);
        }
        let currentDir = root;
        for (let i = dirNames.length - 1; i >= 0; i--) {
            const nextDirName = dirNames[i];
            currentDir = getDirFromDir(currentDir, nextDirName);
        }
        return currentDir;
    }
    function getDirFromDir(dir, name) {
        let target = dir.dirs.get(name);
        if (!target) {
            target = {
                dirs: new Map(),
                fileTexts: new Map(),
                fileTypes: new Map(),
                searched: false,
            };
            dir.dirs.set(name, target);
        }
        return target;
    }
}
exports.createWebFileSystemHost = createWebFileSystemHost;
//# sourceMappingURL=webFileSystemHost.js.map